import csv
from pathlib import Path
from typing import Union, Optional, Dict

from spark_auto_mapper.type_definitions.defined_types import (
    AutoMapperTextInputType,
    AutoMapperAnyDataType,
)


class TranslatorProxyBase:
    def __init__(self, location: Union[str, Path], csv_file: Union[str, Path]) -> None:
        self.location: Union[str, Path] = location
        self.csv_file: Union[str, Path] = csv_file
        self.mapping_dict: Optional[
            Dict[AutoMapperTextInputType, AutoMapperAnyDataType]
        ] = None

    def get_mapping(self) -> Dict[AutoMapperTextInputType, AutoMapperAnyDataType]:
        # read csv
        # convert first two columns to dictionary
        if not self.mapping_dict:
            self.mapping_dict = {}
            with open(
                Path(self.location).joinpath(self.csv_file), newline=""
            ) as csvfile:
                csv_reader = csv.reader(csvfile, delimiter=",", quotechar='"')
                # next(csv_reader)  # skip header
                for row in csv_reader:
                    self.mapping_dict[row[0]] = row[1]
        return self.mapping_dict



import json
from typing import Any

import requests
from requests.adapters import HTTPAdapter


class SlackClient:
    """
    Implements a basic Slack client in http to post messages to a channel
    From https://keestalkstech.com/2019/10/simple-python-code-to-send-message-to-slack-channel-without-packages/
    """

    def __init__(self, slack_token: str, channel: str, bot_user_name: str) -> None:
        self.slack_token = slack_token
        self.slack_channel = channel
        self.slack_icon_url = "https://www.flaticon.com/free-icon/freepik_23317"
        self.slack_user_name = bot_user_name

    def post_message_to_slack(self, text: str, blocks: Any = None) -> Any:
        headers = {
            "Content-type": "application/json; charset=utf-8",
            "Authorization": f"Bearer {self.slack_token}",
        }
        adapter = HTTPAdapter()
        http = requests.Session()
        http.mount("https://", adapter)

        post = http.post(
            "https://slack.com/api/chat.postMessage",
            data=json.dumps(
                {
                    "channel": self.slack_channel,
                    "text": text,
                    "icon_url": self.slack_icon_url,
                    "username": self.slack_user_name,
                    "blocks": json.dumps(blocks) if blocks else None,
                }
            ),
            headers=headers,
        )
        return post.json()



from spark_pipeline_framework.proxy_generator.translator_proxy_base import (
    TranslatorProxyBase,
)
from os import path


# This file was auto-generated by generate_proxies(). It enables auto-complete in PyCharm. Do not edit manually!
class TranslatorsMyTranslator(TranslatorProxyBase):
    def __init__(self) -> None:
        location: str = path.dirname(path.abspath(__file__))
        super().__init__(location=location, csv_file="provider_types.csv")



from os import path, makedirs
from pathlib import Path
from shutil import rmtree

from pyspark.sql import SparkSession, DataFrame
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_csv_loader.v1.framework_csv_loader import (
    FrameworkCsvLoader,
)
from spark_pipeline_framework.transformers.framework_data_frame_analyzer.v1.framework_data_frame_analyzer import (
    FrameworkDataFrameAnalyzer,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)


def test_framework_data_frame_analyzer(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")

    temp_folder = data_dir.joinpath("./temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    test_file_path: str = f"{data_dir.joinpath('test.csv')}"

    df: DataFrame = create_empty_dataframe(spark_session=spark_session)

    FrameworkCsvLoader(
        view="my_view", path_to_csv=test_file_path, delimiter=","
    ).transform(df)

    spark_session.table("my_view").show()

    # Act
    with ProgressLogger() as progress_logger:
        FrameworkDataFrameAnalyzer(
            view="my_view",
            analysis_views_prefix="my_view",
            output_folder=temp_folder,
            progress_logger=progress_logger,
        ).transform(df)

    # Assert
    for column_name in spark_session.table("my_view").columns:
        spark_session.table(f"my_view_{column_name}").show()

    assert (
        spark_session.table("my_view_Column2")
        .where("Column2 == 'foo'")
        .select("count")
        .collect()[0][0]
        == 2
    )



from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger

from spark_pipeline_framework.utilities.slack_client import SlackClient


class SlackProgressLogger(ProgressLogger):
    def __init__(
        self, slack_token: str, slack_channel: str, bot_user_name: str
    ) -> None:
        super().__init__()
        assert slack_token
        assert slack_channel
        assert bot_user_name
        self.slack_token: str = slack_token
        self.slack_channel: str = slack_channel
        self.slack_user_name: str = bot_user_name
        self.slack_client: SlackClient = SlackClient(
            slack_token=slack_token, channel=slack_channel, bot_user_name=bot_user_name
        )

    def log_progress_event(
        self, event_name: str, current: int, total: int, event_format_string: str
    ) -> None:
        self.slack_client.post_message_to_slack(
            event_format_string.format(event_name, current, total)
        )

    def log_event(self, event_name: str, event_text: str) -> None:
        self.slack_client.post_message_to_slack(event_text)



from typing import Dict, Any, Optional

from furl import furl
from requests import Response
from requests.adapters import HTTPAdapter

import requests
from urllib3 import Retry  # type: ignore

# noinspection PyProtectedMember
from pyspark import keyword_only
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FrameworkUrlLoader(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        server_url: str,
        relative_url: str,
        method: str,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        # add a param
        self.server_url: Param = Param(self, "server_url", "")
        self._setDefault(server_url=server_url)

        self.relative_url: Param = Param(self, "relative_url", "")
        self._setDefault(relative_url=relative_url)

        self.method: Param = Param(self, "method", "")
        self._setDefault(method=method)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring, PyUnusedLocal
    @keyword_only
    def setParams(
        self,
        server_url: str,
        relative_url: str,
        method: str,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ) -> Any:
        kwargs = self._input_kwargs
        super().setParams(
            name=name, parameters=parameters, progress_logger=progress_logger
        )
        return self._set(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        server_url: str = self.getServerUrl()
        relative_url: str = self.getRelativeUrl()
        method: str = self.getMethod()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()

        retry_strategy = Retry(
            total=5,
            status_forcelist=[429, 500, 502, 503, 504],
            allowed_methods=[
                "HEAD",
                "GET",
                "PUT",
                "DELETE",
                "OPTIONS",
                "TRACE",
                "POST",
            ],
            backoff_factor=1,
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        http = requests.Session()
        http.mount("https://", adapter)
        http.mount("http://", adapter)

        full_uri: furl = furl(server_url)
        full_uri /= relative_url

        response: Response = http.request(url=full_uri.url, method=method)
        if progress_logger:
            progress_logger.write_to_log(
                f"{method} {full_uri.url} [{response.status_code}]: {str(response.content)}"
            )
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getServerUrl(self) -> str:
        return self.getOrDefault(self.server_url)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getRelativeUrl(self) -> str:
        return self.getOrDefault(self.relative_url)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMethod(self) -> str:
        return self.getOrDefault(self.method)  # type: ignore



import os
from os import listdir, path
from pathlib import Path
from typing import Optional, List, Dict, Any, Union

from pyspark.ml.base import Transformer
from pyspark.sql import DataFrame

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.proxy_generator.python_transformer_helpers import (
    get_python_transformer_from_location,
    get_python_function_from_location,
)
from spark_pipeline_framework.transformers.framework_csv_loader.v1.framework_csv_loader import (
    FrameworkCsvLoader,
)
from spark_pipeline_framework.transformers.framework_mapping_runner.v1.framework_mapping_runner import (
    FrameworkMappingLoader,
)
from spark_pipeline_framework.transformers.framework_sql_transformer.v1.framework_sql_transformer import (
    FrameworkSqlTransformer,
)
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class ProxyBase(FrameworkTransformer):
    def __init__(
        self,
        parameters: Dict[str, Any],
        location: Union[str, Path],
        progress_logger: Optional[ProgressLogger] = None,
        verify_count_remains_same: bool = False,
    ) -> None:
        super().__init__(
            name=self.__class__.__name__,
            parameters=parameters,
            progress_logger=progress_logger,
        )
        self.verify_count_remains_same: bool = verify_count_remains_same
        self.location: str = str(location)
        self.my_transformers: List[Transformer] = []

        assert self.location
        # Iterate over files to create transformers
        files: List[str] = listdir(self.location)
        index_of_module: int = self.location.rfind("/library/")
        module_ = index_of_module + 1
        module_name: str = self.location[module_:].replace("/", ".")

        # noinspection Mypy
        self._set(
            name=self.__class__.__name__,
            parameters=parameters,
            progress_logger=progress_logger,
        )

        for file in files:
            if file.endswith(".csv"):
                file_name = file.replace(".csv", "")
                self.my_transformers.append(
                    FrameworkCsvLoader(
                        view=file_name,
                        path_to_csv=path.join(self.location, file),
                        delimiter=parameters.get("delimiter", ","),
                        has_header=parameters.get("has_header", True),
                    )
                )
            elif file.endswith(".sql"):
                feature_sql: str = self.read_file_as_string(
                    path.join(self.location, file)
                ).format(parameters=parameters)
                self.my_transformers.append(
                    FrameworkSqlTransformer(
                        sql=feature_sql,
                        name=module_name,
                        progress_logger=progress_logger,
                        log_sql=parameters.get("debug_log_sql", False),
                        view=file.replace(".sql", ""),
                        verify_count_remains_same=verify_count_remains_same,
                    )
                )
            elif file.endswith("mapping.py"):
                file_name_only: str = os.path.basename(file)
                # strip off .py to get the module name
                import_module_name: str = file_name_only.replace(".py", "")
                self.my_transformers.append(
                    self.get_python_mapping_transformer("." + import_module_name)
                )
            elif file.endswith("calculate.py") or file.endswith("pipeline.py"):
                file_name_only = file.replace(".py", "")
                self.my_transformers.append(
                    self.get_python_transformer(f".{file_name_only}")
                )

    @staticmethod
    def read_file_as_string(file_path: str) -> str:
        with open(file_path, "r") as file:
            file_contents = file.read()
        return file_contents

    @property
    def transformers(self) -> List[Transformer]:
        return [
            transformer
            for transformer in self.my_transformers
            if transformer is not None
        ]

    @transformers.setter
    def transformers(self, value: Any) -> None:
        raise AttributeError("transformers property is read only.")

    def _transform(self, df: DataFrame) -> DataFrame:
        # iterate through my transformers
        transformer: Transformer
        for transformer in self.my_transformers:
            df = transformer.transform(df)
        return df

    # noinspection PyUnusedLocal
    def fit(self, df: DataFrame) -> Transformer:
        return self

    def get_python_transformer(self, import_module_name: str) -> Transformer:
        progress_logger = self.getProgressLogger()
        assert progress_logger
        return get_python_transformer_from_location(
            location=self.location,
            import_module_name=import_module_name,
            parameters=self.getParameters() or {},
            progress_logger=progress_logger,
        )

    def get_python_mapping_transformer(self, import_module_name: str) -> Transformer:
        parameters: Optional[Dict[str, Any]] = self.getParameters()
        return FrameworkMappingLoader(
            view=parameters["view"] if parameters and "view" in parameters else "",
            mapping_function=get_python_function_from_location(
                location=self.location, import_module_name=import_module_name
            ),
            parameters=parameters,
            progress_logger=self.getProgressLogger(),
        )



from pathlib import Path
from typing import Union, List, Dict, Any, Optional

from pyspark import keyword_only
from pyspark.ml.param.shared import Param
from pyspark.sql.dataframe import DataFrame

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger

from spark_pipeline_framework.transformers.framework_local_file_loader.v1.framework_local_file_loader import (
    FrameworkLocalFileLoader,
)


class FrameworkJsonLoader(FrameworkLocalFileLoader):

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    @keyword_only
    def __init__(
        self,
        view: str,
        filepath: Union[str, List[str], Path],
        clean_column_names: bool = False,
        name: Optional[str] = None,
        progress_logger: Optional[ProgressLogger] = None,
        parameters: Optional[Dict[str, Any]] = None,
        **kwargs: Dict[Any, Any]
    ):
        super().__init__(
            name=name,
            view=view,
            filepath=filepath,
            progress_logger=progress_logger,
            parameters=parameters,
            clean_column_names=clean_column_names,
            **kwargs
        )

        self.multiLine: Param = Param(self, "multiLine", "")
        self._setDefault(multiLine=False)
        self._set(multiLine=False)

    def preprocess(self, df: DataFrame, absolute_paths: List[str]) -> None:
        """
        In pre-processing we try to detect whether the file is a normal json or ndjson
        :param df: DataFrame
        :param absolute_paths: list of paths
        """
        assert absolute_paths
        text_df: DataFrame = df.sql_ctx.read.text(absolute_paths)
        # read the first line of the file
        first_line: str = text_df.select("value").limit(1).collect()[0][0]
        if first_line.lstrip().startswith("["):
            self.setMultiLine(True)
        else:
            self.setMultiLine(False)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setMultiLine(self, value: bool) -> "FrameworkJsonLoader":
        self._paramMap[self.multiLine] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMultiLine(self) -> str:
        return self.getOrDefault(self.multiLine)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getReaderFormat(self) -> str:
        return "json"

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getReaderOptions(self) -> Dict[str, Any]:
        options: Dict[str, Any] = {"multiLine": self.getMultiLine()}
        return options



from typing import Dict, Any, Optional, List, Union

# noinspection PyProtectedMember
from pyspark import keyword_only

# noinspection PyUnresolvedReferences
from pyspark.sql.functions import col
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FrameworkFilterByColumnTransformer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        # add your parameters here (be sure to add them to setParams below too)
        column: str,
        include_only: List[Union[str, int, float]],
        view: str,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        assert column
        self.column: Param = Param(self, "column", "")
        self._setDefault(column=column)

        assert view
        self.view: Param = Param(self, "view", "")
        self._setDefault(view=view)

        assert include_only and isinstance(include_only, list)
        self.include_only: Param = Param(self, "include_only", "")
        self._setDefault(include_only=include_only)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring, PyUnusedLocal
    @keyword_only
    def setParams(
        self,
        # add your parameters here
        column: str,
        include_only: List[Union[str, int, float]],
        view: str,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ) -> Any:
        kwargs = self._input_kwargs
        super().setParams(
            name=name, parameters=parameters, progress_logger=progress_logger
        )
        return self._set(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: str = self.getView()
        column: str = self.getColumn()
        include_only: List[Union[str, int, float]] = self.getIncludeOnly()

        result_df: DataFrame = df.sql_ctx.table(view)
        result_df = result_df.where(col(column).isin(include_only))
        result_df.createOrReplaceTempView(view)

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getColumn(self) -> str:
        return self.getOrDefault(self.column)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getIncludeOnly(self) -> List[Union[str, int, float]]:
        return self.getOrDefault(self.include_only)  # type: ignore



from pathlib import Path

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from pyspark.sql.types import StructType

from library.pipelines.my_pipeline.v1.pipelines_my_pipeline_v1 import (
    PipelinesMyPipelineV1,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger


def test_can_run_framework_pipeline_transformer(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")
    flights_path: str = f"file://{data_dir.joinpath('flights.csv')}"

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    spark_session.sql("DROP TABLE IF EXISTS default.flights")

    # Act
    parameters = {"flights_path": flights_path}

    with ProgressLogger() as progress_logger:
        pipeline: PipelinesMyPipelineV1 = PipelinesMyPipelineV1(
            parameters=parameters, progress_logger=progress_logger
        )
        transformer = pipeline.fit(df)
        transformer.transform(df)

    # Assert
    result_df: DataFrame = spark_session.sql("SELECT * FROM flights2")
    result_df.show()

    assert result_df.count() > 0
